@@ -18,25 +18,6 @@
  */
 package org.apache.felix.framework;
 
-import java.io.File;
-import java.io.IOException;
-import java.io.InputStream;
-import java.net.URL;
-import java.security.AccessControlContext;
-import java.security.ProtectionDomain;
-import java.util.ArrayList;
-import java.util.Collection;
-import java.util.Dictionary;
-import java.util.Enumeration;
-import java.util.HashMap;
-import java.util.Iterator;
-import java.util.List;
-import java.util.Locale;
-import java.util.Map;
-import java.util.Properties;
-import java.util.Set;
-import java.util.StringTokenizer;
-
 import org.apache.felix.framework.cache.BundleArchive;
 import org.apache.felix.framework.util.SecurityManagerEx;
 import org.apache.felix.framework.util.ShrinkableCollection;
@@ -60,13 +41,33 @@ import org.osgi.framework.wiring.BundleRevisions;
 import org.osgi.framework.wiring.BundleWire;
 import org.osgi.framework.wiring.BundleWiring;
 
+import java.io.File;
+import java.io.IOException;
+import java.io.InputStream;
+import java.net.URL;
+import java.security.AccessControlContext;
+import java.security.ProtectionDomain;
+import java.util.ArrayList;
+import java.util.Collection;
+import java.util.Dictionary;
+import java.util.Enumeration;
+import java.util.HashMap;
+import java.util.Iterator;
+import java.util.List;
+import java.util.Locale;
+import java.util.Map;
+import java.util.Properties;
+import java.util.Set;
+import java.util.StringTokenizer;
+
 class BundleImpl implements Bundle, BundleRevisions
 {
     // No one should use this field directly, use getFramework() instead.
     private final Felix __m_felix;
 
     private final BundleArchive m_archive;
-    private final List<BundleRevision> m_revisions = new ArrayList<BundleRevision>(0);
+    private final List<BundleRevisionImpl> m_revisions = new ArrayList<BundleRevisionImpl>(0);
+    private volatile BundleRevisionImpl m_currentRevision = null;
     private volatile int m_state;
     private boolean m_useDeclaredActivationPolicy;
     private BundleActivator m_activator = null;
@@ -110,7 +111,7 @@ class BundleImpl implements Bundle, BundleRevisions
         m_context = null;
         m_installingBundle = installingBundle;
 
-        BundleRevision revision = createRevision(false);
+        BundleRevisionImpl revision = createRevision(false);
         addRevision(revision);
     }
 
@@ -166,13 +167,13 @@ class BundleImpl implements Bundle, BundleRevisions
     {
         // Remove the bundle's associated revisions from the resolver state
         // and close them.
-        for (BundleRevision br : m_revisions)
+        for (BundleRevisionImpl br : m_revisions)
         {
             // Remove the revision from the resolver state.
             getFramework().getResolver().removeRevision(br);
 
             // Close the revision's content.
-            ((BundleRevisionImpl) br).close();
+            br.close();
         }
     }
 
@@ -188,16 +189,24 @@ class BundleImpl implements Bundle, BundleRevisions
         {
             // Get current revision, since we can reuse it.
             BundleRevisionImpl current = adapt(BundleRevisionImpl.class);
-            // Close all existing revisions.
-            closeRevisions();
-            // Clear all revisions.
+
+            if (isRemovalPending()) {
+                closeRevisions();
+                // Purge all old archive revisions, only keeping the newest one.
+                m_archive.purge();
+
+                current.resetContent(m_archive.getCurrentRevision().getContent());
+            }
+            else {
+                // Remove the revision from the resolver state.
+                getFramework().getResolver().removeRevision(current);
+                current.resolve(null);
+                current.disposeContentPath();
+            }
+
             m_revisions.clear();
+            m_currentRevision = null;
 
-            // Purge all old archive revisions, only keeping the newest one.
-            m_archive.purge();
-
-            // Reset the content of the current bundle revision.
-            current.resetContent(m_archive.getCurrentRevision().getContent());
             // Re-add the revision to the bundle.
             addRevision(current);
 
@@ -233,6 +242,7 @@ class BundleImpl implements Bundle, BundleRevisions
         m_activator = activator;
     }
 
+    @Override
     public BundleContext getBundleContext()
     {
         Object sm = System.getSecurityManager();
@@ -251,6 +261,7 @@ class BundleImpl implements Bundle, BundleRevisions
         m_context = context;
     }
 
+    @Override
     public long getBundleId()
     {
         try
@@ -268,6 +279,7 @@ class BundleImpl implements Bundle, BundleRevisions
         }
     }
 
+    @Override
     public URL getEntry(String name)
     {
         Object sm = System.getSecurityManager();
@@ -288,6 +300,7 @@ class BundleImpl implements Bundle, BundleRevisions
         return getFramework().getBundleEntry(this, name);
     }
 
+    @Override
     public Enumeration getEntryPaths(String path)
     {
         Object sm = System.getSecurityManager();
@@ -308,6 +321,7 @@ class BundleImpl implements Bundle, BundleRevisions
         return getFramework().getBundleEntryPaths(this, path);
     }
 
+    @Override
     public Enumeration findEntries(String path, String filePattern, boolean recurse)
     {
         Object sm = System.getSecurityManager();
@@ -329,11 +343,13 @@ class BundleImpl implements Bundle, BundleRevisions
                 this, path, filePattern, recurse);
     }
 
+    @Override
     public Dictionary getHeaders()
     {
         return getHeaders(Locale.getDefault().toString());
     }
 
+    @Override
     public Dictionary getHeaders(String locale)
     {
         Object sm = System.getSecurityManager();
@@ -553,7 +569,7 @@ class BundleImpl implements Bundle, BundleRevisions
         List<String> result = new ArrayList(4);
 
         StringTokenizer tokens;
-        StringBuffer tempLocale = new StringBuffer(basename);
+        StringBuilder tempLocale = new StringBuilder(basename);
 
         result.add(tempLocale.toString());
 
@@ -569,6 +585,7 @@ class BundleImpl implements Bundle, BundleRevisions
         return result;
     }
 
+    @Override
     public long getLastModified()
     {
         try
@@ -602,6 +619,7 @@ class BundleImpl implements Bundle, BundleRevisions
         }
     }
 
+    @Override
     public String getLocation()
     {
         Object sm = System.getSecurityManager();
@@ -636,6 +654,7 @@ class BundleImpl implements Bundle, BundleRevisions
      *
      * @return a URL to named resource, or null if not found.
     **/
+    @Override
     public URL getResource(String name)
     {
         Object sm = System.getSecurityManager();
@@ -656,6 +675,7 @@ class BundleImpl implements Bundle, BundleRevisions
         return getFramework().getBundleResource(this, name);
     }
 
+    @Override
     public Enumeration getResources(String name) throws IOException
     {
         Object sm = System.getSecurityManager();
@@ -685,6 +705,7 @@ class BundleImpl implements Bundle, BundleRevisions
      *
      * @return an array of service references or null.
     **/
+    @Override
     public ServiceReference[] getRegisteredServices()
     {
         Object sm = System.getSecurityManager();
@@ -728,6 +749,7 @@ class BundleImpl implements Bundle, BundleRevisions
         }
     }
 
+    @Override
     public ServiceReference[] getServicesInUse()
     {
         Object sm = System.getSecurityManager();
@@ -769,6 +791,7 @@ class BundleImpl implements Bundle, BundleRevisions
         return getFramework().getBundleServicesInUse(this);
     }
 
+    @Override
     public int getState()
     {
         return m_state;
@@ -903,9 +926,9 @@ class BundleImpl implements Bundle, BundleRevisions
 
     synchronized boolean isExtension()
     {
-        for (BundleRevision revision : m_revisions)
+        for (BundleRevisionImpl revision : m_revisions)
         {
-            if (((BundleRevisionImpl) revision).isExtension())
+            if (revision.isExtension())
             {
                 return true;
             }
@@ -913,27 +936,32 @@ class BundleImpl implements Bundle, BundleRevisions
         return false;
     }
 
+    @Override
     public String getSymbolicName()
     {
         return adapt(BundleRevisionImpl.class).getSymbolicName();
     }
 
+    @Override
     public Version getVersion()
     {
         return adapt(BundleRevisionImpl.class).getVersion();
     }
 
+    @Override
     public boolean hasPermission(Object obj)
     {
         return getFramework().bundleHasPermission(this, obj);
     }
 
+    @Override
     public Map getSignerCertificates(int signersType)
     {
         // TODO: SECURITY - This needs to be adapted to our security mechanisms.
         return (Map) getFramework().getSignerMatcher(this, signersType);
     }
 
+    @Override
     public Class loadClass(String name) throws ClassNotFoundException
     {
         if (isExtension())
@@ -959,11 +987,13 @@ class BundleImpl implements Bundle, BundleRevisions
         return getFramework().loadBundleClass(this, name);
     }
 
+    @Override
     public void start() throws BundleException
     {
         start(0);
     }
 
+    @Override
     public void start(int options) throws BundleException
     {
         Object sm = System.getSecurityManager();
@@ -977,11 +1007,13 @@ class BundleImpl implements Bundle, BundleRevisions
         getFramework().startBundle(this, options);
     }
 
+    @Override
     public void update() throws BundleException
     {
         update(null);
     }
 
+    @Override
     public void update(InputStream is) throws BundleException
     {
         Object sm = System.getSecurityManager();
@@ -995,11 +1027,13 @@ class BundleImpl implements Bundle, BundleRevisions
         getFramework().updateBundle(this, is);
     }
 
+    @Override
     public void stop() throws BundleException
     {
         stop(0);
     }
 
+    @Override
     public void stop(int options) throws BundleException
     {
         Object sm = System.getSecurityManager();
@@ -1013,6 +1047,7 @@ class BundleImpl implements Bundle, BundleRevisions
         getFramework().stopBundle(this, ((options & Bundle.STOP_TRANSIENT) == 0));
     }
 
+    @Override
     public void uninstall() throws BundleException
     {
         Object sm = System.getSecurityManager();
@@ -1054,9 +1089,9 @@ class BundleImpl implements Bundle, BundleRevisions
         Object sm = System.getSecurityManager();
         if ((sm != null) && (getFramework().getSecurityProvider() != null))
         {
-            Class caller = m_smEx.getClassContext()[3];
-            if (((Felix.m_secureAction.getClassLoader(caller) != m_classloader) ||
-                !caller.getName().startsWith("org.apache.felix.framework.")))
+            Class[] classes = m_smEx.getClassContext();
+            if (classes.length < 3 || ((Felix.m_secureAction.getClassLoader(classes[3]) != m_classloader) ||
+                !classes[3].getName().startsWith("org.apache.felix.framework.")))
             {
                 ((SecurityManager) sm).checkPermission(
                     new AdaptPermission(type.getName(), this, AdaptPermission.ADAPT));
@@ -1064,7 +1099,8 @@ class BundleImpl implements Bundle, BundleRevisions
         }
     }
 
-    public synchronized <A> A adapt(Class<A> type)
+    @Override
+    public <A> A adapt(Class<A> type)
     {
         checkAdapt(type);
         if (type == BundleContext.class)
@@ -1082,14 +1118,14 @@ class BundleImpl implements Bundle, BundleRevisions
             {
                 return null;
             }
-            return (A) m_revisions.get(0);
+            return (A) m_currentRevision;
         }
         // We need some way to get the current revision even if
         // the associated bundle is uninstalled, so we use the
         // impl revision class for this purpose.
         else if (type == BundleRevisionImpl.class)
         {
-            return (A) m_revisions.get(0);
+            return (A) m_currentRevision;
         }
         else if (type == BundleRevisions.class)
         {
@@ -1101,7 +1137,9 @@ class BundleImpl implements Bundle, BundleRevisions
             {
                 return null;
             }
-            return (A) m_revisions.get(0).getWiring();
+            BundleRevisionImpl revision = m_currentRevision;
+
+            return (A) (revision != null ? revision.getWiring() : null);
         }
         else if ( type == AccessControlContext.class)
         {
@@ -1125,11 +1163,13 @@ class BundleImpl implements Bundle, BundleRevisions
         return null;
     }
 
+    @Override
     public File getDataFile(String filename)
     {
         return getFramework().getDataFile(this, filename);
     }
 
+    @Override
     public int compareTo(Bundle t)
     {
         long thisBundleId = this.getBundleId();
@@ -1157,11 +1197,13 @@ class BundleImpl implements Bundle, BundleRevisions
     // Revision management.
     //
 
+    @Override
     public Bundle getBundle()
     {
         return this;
     }
 
+    @Override
     public synchronized List<BundleRevision> getRevisions()
     {
         return new ArrayList<BundleRevision>(m_revisions);
@@ -1185,7 +1227,7 @@ class BundleImpl implements Bundle, BundleRevisions
         m_archive.revise(location, is);
         try
         {
-            BundleRevision revision = createRevision(true);
+            BundleRevisionImpl revision = createRevision(true);
             addRevision(revision);
         }
         catch (Exception ex)
@@ -1197,14 +1239,13 @@ class BundleImpl implements Bundle, BundleRevisions
 
     synchronized boolean rollbackRevise() throws Exception
     {
-        boolean isExtension = isExtension();
         BundleRevision br = m_revisions.remove(0);
-        if (!isExtension)
-        {
-            // Since revising a bundle adds a revision to the global
-            // state, we must remove it from the global state on rollback.
-            getFramework().getResolver().removeRevision(br);
-        }
+        m_currentRevision = !m_revisions.isEmpty() ? m_revisions.get(0) : null;
+
+        // Since revising a bundle adds a revision to the global
+        // state, we must remove it from the global state on rollback.
+        getFramework().getResolver().removeRevision(br);
+
         return m_archive.rollbackRevise();
     }
 
@@ -1212,35 +1253,34 @@ class BundleImpl implements Bundle, BundleRevisions
     // system bundle needs to add its revision directly to the bundle,
     // since it doesn't have an archive from which it will be created,
     // which is the normal case.
-    synchronized void addRevision(BundleRevision revision) throws Exception
+    synchronized void addRevision(BundleRevisionImpl revision) throws Exception
     {
+        BundleRevisionImpl previous = m_currentRevision;
         m_revisions.add(0, revision);
+        m_currentRevision = revision;
 
         try
         {
-            getFramework().setBundleProtectionDomain(this, (BundleRevisionImpl) revision);
+            getFramework().setBundleProtectionDomain(revision);
         }
         catch (Exception ex)
         {
             m_revisions.remove(0);
+            m_currentRevision = previous;
             throw ex;
         }
 
-        // TODO: REFACTOR - consider nulling capabilities for extension bundles
-        // so we don't need this check anymore.
-        if (!isExtension())
-        {
-            // Now that the revision is added to the bundle, we can update
-            // the resolver's state to be aware of any new capabilities.
-            getFramework().getResolver().addRevision(revision);
-        }
+        /// Now that the revision is added to the bundle, we can update
+        // the resolver's state to be aware of any new capabilities.
+        getFramework().getResolver().addRevision(revision);
     }
 
-    private BundleRevision createRevision(boolean isUpdate) throws Exception
+    private BundleRevisionImpl createRevision(boolean isUpdate) throws Exception
     {
         // Get and parse the manifest from the most recent revision and
         // create an associated revision object for it.
-        Map headerMap = m_archive.getCurrentRevision().getManifestHeader();
+        Map headerMap = Util.getMultiReleaseAwareManifestHeaders(
+            getFramework()._getProperty("java.specification.version"), m_archive.getCurrentRevision());
 
         // Create the bundle revision instance.
         BundleRevisionImpl revision = new BundleRevisionImpl(
@@ -1280,7 +1320,7 @@ class BundleImpl implements Bundle, BundleRevisions
             }
             if (!collisionCanditates.isEmpty() && allowMultiple.equals(Constants.FRAMEWORK_BSNVERSION_MANAGED))
             {
-                Set<ServiceReference<CollisionHook>> hooks = getFramework().getHooks(CollisionHook.class);
+                Set<ServiceReference<CollisionHook>> hooks = getFramework().getHookRegistry().getHooks(CollisionHook.class);
                 if (!hooks.isEmpty())
                 {
                     Collection<Bundle> shrinkableCollisionCandidates = new ShrinkableCollection<Bundle>(collisionCanditates);
@@ -1308,7 +1348,7 @@ class BundleImpl implements Bundle, BundleRevisions
                     }
                 }
             }
-            if (!collisionCanditates.isEmpty())
+            if (!collisionCanditates.isEmpty() && m_installingBundle != null)
             {
                 throw new BundleException(
                     "Bundle symbolic name and version are not unique: "
@@ -1323,9 +1363,9 @@ class BundleImpl implements Bundle, BundleRevisions
     {
         ProtectionDomain pd = null;
 
-        for (int i = m_revisions.size() - 1; (i >= 0) && (pd == null); i--)
+        for (int i = 0; (i < m_revisions.size()) && (pd == null); i++)
         {
-            pd = ((BundleRevisionImpl) m_revisions.get(i)).getProtectionDomain();
+            pd = m_revisions.get(i).getProtectionDomain();
         }
 
         return pd;
