@@ -1,4 +1,4 @@
-/* 
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -18,21 +18,24 @@
  */
 package org.apache.felix.framework;
 
+import java.lang.reflect.Method;
 import java.net.ContentHandler;
 import java.net.ContentHandlerFactory;
 import java.net.URL;
 import java.net.URLConnection;
 import java.net.URLStreamHandler;
 import java.net.URLStreamHandlerFactory;
-import java.util.ArrayList;
-import java.util.HashMap;
 import java.util.List;
-import java.util.Map;
 import java.util.StringTokenizer;
+import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.CopyOnWriteArrayList;
+
+import static org.apache.felix.framework.util.Util.putIfAbsentAndReturn;
 
 import org.apache.felix.framework.util.FelixConstants;
 import org.apache.felix.framework.util.SecureAction;
 import org.apache.felix.framework.util.SecurityManagerEx;
+import org.osgi.framework.Constants;
 import org.osgi.service.url.URLStreamHandlerService;
 
 /**
@@ -85,62 +88,73 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
     private static volatile SecurityManagerEx m_sm = null;
     private static volatile URLHandlers m_handler = null;
 
-    // This maps classloaders of URLHandlers in other classloaders to lists of 
+    // This maps classloaders of URLHandlers in other classloaders to lists of
     // their frameworks.
-    private static Map m_classloaderToFrameworkLists = new HashMap();
+    private final static ConcurrentHashMap<ClassLoader, List<Object>> m_classloaderToFrameworkLists = new ConcurrentHashMap<ClassLoader, List<Object>>();
 
-    // The list to hold all enabled frameworks registered with this handlers 
-    private static final List m_frameworks = new ArrayList();
-    private static int m_counter = 0;
+    // The list to hold all enabled frameworks registered with this handlers
+    private static final CopyOnWriteArrayList<Felix> m_frameworks = new CopyOnWriteArrayList<Felix>();
+    private static volatile int m_counter = 0;
 
-    private static Map m_contentHandlerCache = null;
-    private static Map m_streamHandlerCache = null;
-    private static URLStreamHandlerFactory m_streamHandlerFactory;
-    private static ContentHandlerFactory m_contentHandlerFactory;
+    private static final ConcurrentHashMap<String, ContentHandler> m_contentHandlerCache = new ConcurrentHashMap<String, ContentHandler>();
+    private static final ConcurrentHashMap<String, URLStreamHandler> m_streamHandlerCache = new ConcurrentHashMap<String, URLStreamHandler>();
+    private static final ConcurrentHashMap<String, URL> m_protocolToURL = new ConcurrentHashMap<String, URL>();
+
+    private static volatile URLStreamHandlerFactory m_streamHandlerFactory;
+    private static volatile ContentHandlerFactory m_contentHandlerFactory;
     private static final String STREAM_HANDLER_PACKAGE_PROP = "java.protocol.handler.pkgs";
     private static final String DEFAULT_STREAM_HANDLER_PACKAGE = "sun.net.www.protocol|com.ibm.oti.net.www.protocol|gnu.java.net.protocol|wonka.net|com.acunia.wonka.net|org.apache.harmony.luni.internal.net.www.protocol|weblogic.utils|weblogic.net|javax.net.ssl|COM.newmonics.www.protocols";
-    private static Object m_rootURLHandlers;
+    private static volatile Object m_rootURLHandlers;
 
     private static final String m_streamPkgs;
-    private static final Map m_builtIn = new HashMap();
+    private static final ConcurrentHashMap<String, URLStreamHandler> m_builtIn = new ConcurrentHashMap<String, URLStreamHandler>();
     private static final boolean m_loaded;
-    
-    static 
+
+    static
     {
         String pkgs = new SecureAction().getSystemProperty(STREAM_HANDLER_PACKAGE_PROP, "");
         m_streamPkgs = (pkgs.equals(""))
             ? DEFAULT_STREAM_HANDLER_PACKAGE
             : pkgs + "|" + DEFAULT_STREAM_HANDLER_PACKAGE;
         m_loaded = (null != URLHandlersStreamHandlerProxy.class) &&
-            (null != URLHandlersContentHandlerProxy.class) && (null != URLStreamHandlerService.class);
+            (null != URLHandlersContentHandlerProxy.class) && (null != URLStreamHandlerService.class) && new URLHandlersStreamHandlerProxy(null, null) != null;
     }
 
-    
-    private static final Map m_handlerToURL = new HashMap();
     private void init(String protocol, URLStreamHandlerFactory factory)
     {
         try
         {
-            URLStreamHandler handler = getBuiltInStreamHandler(protocol, factory);
-            if (handler != null)
-            {
-                URL url = new URL(protocol, null, -1, "", handler);
-                m_handlerToURL.put(handler, url);
-            }
+            // Try to get it directly from the URL class to if possible
+            Method getURLStreamHandler = m_secureAction.getDeclaredMethod(URL.class,"getURLStreamHandler", new Class[]{String.class});
+            URLStreamHandler handler = (URLStreamHandler) m_secureAction.invoke(getURLStreamHandler, null, new Object[]{protocol});
+            addToCache(m_builtIn, protocol, handler);
         }
         catch (Throwable ex)
         {
-            // Ignore, this is a best effort (maybe log it or something).
+            // Ignore, this is a best effort
+            try
+            {
+                URLStreamHandler handler = getBuiltInStreamHandler(protocol, factory);
+                if (handler != null)
+                {
+                    URL url = new URL(protocol, null, -1, "", handler);
+                    addToCache(m_protocolToURL, protocol, url);
+                }
+            }
+            catch (Throwable ex2)
+            {
+                // Ignore, this is a best effort (maybe log it or something).
+            }
         }
     }
-    
+
     /**
      * <p>
-     * Only one instance of this class is created per classloader 
+     * Only one instance of this class is created per classloader
      * and that one instance is registered as the stream and content handler
      * factories for the JVM. Unless, we already register one from a different
      * classloader. In this case we attach to this root.
-     * </p> 
+     * </p>
     **/
     private URLHandlers()
     {
@@ -153,36 +167,77 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                 currentFactory = (URLStreamHandlerFactory) m_secureAction.swapStaticFieldIfNotClass(URL.class,
                     URLStreamHandlerFactory.class, URLHANDLERS_CLASS, "streamHandlerLock");
             }
-            catch (Exception ex)
-            {
-                // Ignore, this is a best effort (maybe log it or something)
-            }
-    
-            init("file", currentFactory);
-            init("ftp", currentFactory);
-            init("http", currentFactory);
-            init("https", currentFactory);
-            try
-            {
-                getBuiltInStreamHandler("jar", currentFactory);
-            }
             catch (Throwable ex)
             {
                 // Ignore, this is a best effort (maybe log it or something)
             }
-    
+
+            init("file", currentFactory);
+            init("ftp", currentFactory);
+            init("http", currentFactory);
+            init("https", currentFactory);
+
+
+            // Try to preload the jrt handler as we need it from the jvm on java > 8
+            if (getFromCache(m_builtIn, "jrt") == null)
+            {
+                try
+                {
+                    // Try to get it directly from the URL class to if possible
+                    Method getURLStreamHandler = m_secureAction.getDeclaredMethod(URL.class,"getURLStreamHandler", new Class[]{String.class});
+                    URLStreamHandler handler = (URLStreamHandler) m_secureAction.invoke(getURLStreamHandler, null, new Object[]{"jrt"});
+                    addToCache(m_builtIn, "jrt", handler);
+                }
+                catch (Throwable ex)
+                {
+                    // Ignore, this is a best effort and try to load the normal way
+                    try
+                    {
+                        getBuiltInStreamHandler("jrt", currentFactory);
+                    }
+                    catch (Throwable ex2)
+                    {
+                        // Ignore, this is a best efforts
+                    }
+                }
+            }
+
+            // Try to preload the jrt handler as we need it from the jvm on java > 8
+            if (getFromCache(m_builtIn, "jar") == null)
+            {
+                try
+                {
+                    // Try to get it directly from the URL class to if possible
+                    Method getURLStreamHandler = m_secureAction.getDeclaredMethod(URL.class,"getURLStreamHandler", new Class[]{String.class});
+                    URLStreamHandler handler = (URLStreamHandler) m_secureAction.invoke(getURLStreamHandler, null, new Object[]{"jar"});
+                    addToCache(m_builtIn, "jar", handler);
+                }
+                catch (Throwable ex)
+                {
+                    // Ignore, this is a best effort
+                    try
+                    {
+                        getBuiltInStreamHandler("jar", currentFactory);
+                    }
+                    catch (Throwable ex2)
+                    {
+                        // Ignore, this is a best effort (maybe log it or something)
+                    }
+                }
+            }
+
             if (currentFactory != null)
             {
                 try
                 {
                     URL.setURLStreamHandlerFactory(currentFactory);
                 }
-                catch (Exception ex)
+                catch (Throwable ex)
                 {
                     // Ignore, this is a best effort (maybe log it or something)
                 }
             }
-            
+
             try
             {
                 URL.setURLStreamHandlerFactory(this);
@@ -204,9 +259,9 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                 {
                     // there already is a factory set so try to swap it with ours.
                     m_streamHandlerFactory = (URLStreamHandlerFactory)
-                        m_secureAction.swapStaticFieldIfNotClass(URL.class, 
+                        m_secureAction.swapStaticFieldIfNotClass(URL.class,
                         URLStreamHandlerFactory.class, URLHANDLERS_CLASS, "streamHandlerLock");
-                    
+
                     if (m_streamHandlerFactory == null)
                     {
                         throw err;
@@ -221,16 +276,16 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                         try
                         {
                             m_secureAction.invoke(
-                                m_secureAction.getDeclaredMethod(m_streamHandlerFactory.getClass(), 
-                                "registerFrameworkListsForContextSearch", 
-                                new Class[]{ClassLoader.class, List.class}), 
-                                m_streamHandlerFactory, new Object[]{ URLHANDLERS_CLASS.getClassLoader(), 
+                                m_secureAction.getDeclaredMethod(m_streamHandlerFactory.getClass(),
+                                "registerFrameworkListsForContextSearch",
+                                new Class[]{ClassLoader.class, List.class}),
+                                m_streamHandlerFactory, new Object[]{ URLHANDLERS_CLASS.getClassLoader(),
                                     m_frameworks });
                             m_rootURLHandlers = m_streamHandlerFactory;
                         }
                         catch (Exception ex)
                         {
-                            new RuntimeException(ex.getMessage());
+                            throw new RuntimeException(ex.getMessage());
                         }
                     }
                 }
@@ -239,7 +294,7 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                     throw err;
                 }
             }
-            
+
             try
             {
                 URLConnection.setContentHandlerFactory(this);
@@ -258,10 +313,10 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
             {
                 // there already is a factory set so try to swap it with ours.
                 try
-                {   
-                    m_contentHandlerFactory = (ContentHandlerFactory) 
+                {
+                    m_contentHandlerFactory = (ContentHandlerFactory)
                         m_secureAction.swapStaticFieldIfNotClass(
-                            URLConnection.class, ContentHandlerFactory.class, 
+                            URLConnection.class, ContentHandlerFactory.class,
                             URLHANDLERS_CLASS, null);
                     if (m_contentHandlerFactory == null)
                     {
@@ -284,12 +339,12 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
             m_streamHandlerFactory.getClass().getName())))
         {
             m_sm = null;
-            m_handlerToURL.clear();
+            m_protocolToURL.clear();
             m_builtIn.clear();
         }
     }
 
-    static void registerFrameworkListsForContextSearch(ClassLoader index, 
+    static void registerFrameworkListsForContextSearch(ClassLoader index,
         List frameworkLists)
     {
         synchronized (URL.class)
@@ -316,7 +371,7 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                         {
                             try
                             {
-                                m_secureAction.swapStaticFieldIfNotClass(URL.class, 
+                                m_secureAction.swapStaticFieldIfNotClass(URL.class,
                                     URLStreamHandlerFactory.class, null, "streamHandlerLock");
                             }
                             catch (Exception ex)
@@ -324,7 +379,7 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                                 // TODO log this
                                 ex.printStackTrace();
                             }
-                            
+
                             if (m_streamHandlerFactory.getClass() != URLHANDLERS_CLASS)
                             {
                                 URL.setURLStreamHandlerFactory(m_streamHandlerFactory);
@@ -332,7 +387,7 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                             try
                             {
                                 m_secureAction.swapStaticFieldIfNotClass(
-                                    URLConnection.class, ContentHandlerFactory.class, 
+                                    URLConnection.class, ContentHandlerFactory.class,
                                     null, null);
                             }
                             catch (Exception ex)
@@ -340,7 +395,7 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                                 // TODO log this
                                 ex.printStackTrace();
                             }
-                            
+
                             if (m_contentHandlerFactory.getClass() != URLHANDLERS_CLASS)
                             {
                                 URLConnection.setContentHandlerFactory(m_contentHandlerFactory);
@@ -354,29 +409,31 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
 
     private URLStreamHandler getBuiltInStreamHandler(String protocol, URLStreamHandlerFactory factory)
     {
-        synchronized (m_builtIn)
+        URLStreamHandler handler = getFromCache(m_builtIn, protocol);
+
+        if (handler != null)
         {
-            if (m_builtIn.containsKey(protocol))
-            {
-                return (URLStreamHandler) m_builtIn.get(protocol);
-            }
+            return handler;
         }
+
         if (factory != null)
         {
-            URLStreamHandler result = factory.createURLStreamHandler(protocol);
-            if (result != null)
-            {
-                return addToCache(protocol, result);
-            }
+            handler = factory.createURLStreamHandler(protocol);
         }
-        // Check for built-in handlers for the mime type.
-        // Iterate over built-in packages.
-        URLStreamHandler handler = loadBuiltInStreamHandler(protocol, null);
+
+        if (handler == null)
+        {
+            // Check for built-in handlers for the mime type.
+            // Iterate over built-in packages.
+            handler = loadBuiltInStreamHandler(protocol, null);
+        }
+
         if (handler == null)
         {
             handler = loadBuiltInStreamHandler(protocol, ClassLoader.getSystemClassLoader());
         }
-        return addToCache(protocol, handler);
+
+        return addToCache(m_builtIn, protocol, handler);
     }
 
     private URLStreamHandler loadBuiltInStreamHandler(String protocol, ClassLoader classLoader) {
@@ -388,7 +445,48 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
             try
             {
                 // If a built-in handler is found then cache and return it
-                Class handler = m_secureAction.forName(className, classLoader); 
+                Class handler = m_secureAction.forName(className, classLoader);
+                if (handler != null)
+                {
+                    return (URLStreamHandler) handler.newInstance();
+                }
+            }
+            catch (Throwable ex)
+            {
+                // This could be a class not found exception or an
+                // instantiation exception, not much we can do in either
+                // case other than ignore it.
+            }
+        }
+        // This is a workaround for android - Starting with 4.1 the built-in core handler
+        // are not following the normal naming nore package schema :-(
+        String androidHandler = null;
+        if ("file".equalsIgnoreCase(protocol))
+        {
+            androidHandler = "libcore.net.url.FileHandler";
+        }
+        else if ("ftp".equalsIgnoreCase(protocol))
+        {
+            androidHandler = "libcore.net.url.FtpHandler";
+        }
+        else if ("http".equalsIgnoreCase(protocol))
+        {
+            androidHandler = "libcore.net.http.HttpHandler";
+        }
+        else if ("https".equalsIgnoreCase(protocol))
+        {
+            androidHandler = "libcore.net.http.HttpsHandler";
+        }
+        else if ("jar".equalsIgnoreCase(protocol))
+        {
+            androidHandler = "libcore.net.url.JarHandler";
+        }
+        if (androidHandler != null)
+        {
+            try
+            {
+                // If a built-in handler is found then cache and return it
+                Class handler = m_secureAction.forName(androidHandler, classLoader);
                 if (handler != null)
                 {
                     return (URLStreamHandler) handler.newInstance();
@@ -404,16 +502,6 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
         return null;
     }
 
-    private synchronized URLStreamHandler addToCache(String protocol, URLStreamHandler result)
-    {
-        if (!m_builtIn.containsKey(protocol))
-            {
-                m_builtIn.put(protocol, result);
-                return result;
-            }
-        return (URLStreamHandler) m_builtIn.get(protocol);
-    }
-    
     /**
      * <p>
      * This is a method implementation for the <tt>URLStreamHandlerFactory</tt>
@@ -432,8 +520,8 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
         // performed for code consistency between stream and content
         // handlers and also because caching behavior may not be guaranteed
         // across different JRE implementations.
-        URLStreamHandler handler = getFromStreamCache(protocol);
-        
+        URLStreamHandler handler = getFromCache(m_streamHandlerCache, protocol);
+
         if (handler != null)
         {
             return handler;
@@ -443,17 +531,16 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
         // allowed to deal with it.
         if (protocol.equals(FelixConstants.BUNDLE_URL_PROTOCOL))
         {
-            return addToStreamCache(protocol, 
-                new URLHandlersBundleStreamHandler(m_secureAction));
+            return new URLHandlersBundleStreamHandler(getFrameworkFromContext(), m_secureAction);
         }
 
-       handler = getBuiltInStreamHandler(protocol, 
-           (m_streamHandlerFactory != this) ? m_streamHandlerFactory : null);
+        handler = getBuiltInStreamHandler(protocol,
+            (m_streamHandlerFactory != this) ? m_streamHandlerFactory : null);
 
         // If built-in content handler, then create a proxy handler.
-        return addToStreamCache(protocol, 
-            new URLHandlersStreamHandlerProxy(protocol, m_secureAction, 
-                handler, (URL) m_handlerToURL.get(handler)));
+        return addToCache(m_streamHandlerCache, protocol,
+            new URLHandlersStreamHandlerProxy(protocol, m_secureAction,
+                handler, getFromCache(m_protocolToURL, protocol)));
     }
 
     /**
@@ -474,63 +561,26 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
         // performed for code consistency between stream and content
         // handlers and also because caching behavior may not be guaranteed
         // across different JRE implementations.
-        ContentHandler handler = getFromContentCache(mimeType);
-        
+        ContentHandler handler = getFromCache(m_contentHandlerCache, mimeType);
+
         if (handler != null)
         {
             return handler;
         }
 
-        return addToContentCache(mimeType, 
-            new URLHandlersContentHandlerProxy(mimeType, m_secureAction, 
+        return addToCache(m_contentHandlerCache, mimeType,
+            new URLHandlersContentHandlerProxy(mimeType, m_secureAction,
             (m_contentHandlerFactory != this) ? m_contentHandlerFactory : null));
     }
 
-    private synchronized ContentHandler addToContentCache(String mimeType, ContentHandler handler)
+    private static <K,V> V addToCache(ConcurrentHashMap<K,V> cache, K key, V value)
     {
-        if (m_contentHandlerCache == null)
-        {
-            m_contentHandlerCache = new HashMap();
-        }
-        return (ContentHandler) addToCache(m_contentHandlerCache, mimeType, handler);
-    }
-    
-    private synchronized ContentHandler getFromContentCache(String mimeType)
-    {
-        return (ContentHandler) ((m_contentHandlerCache != null) ? 
-            m_contentHandlerCache.get(mimeType) : null);
+        return key != null && value != null ? putIfAbsentAndReturn(cache, key, value) : null;
     }
 
-    private synchronized URLStreamHandler addToStreamCache(String protocol, URLStreamHandler handler)
+    private static <K,V> V getFromCache(ConcurrentHashMap<K,V> cache, K key)
     {
-        if (m_streamHandlerCache == null)
-        {
-            m_streamHandlerCache = new HashMap();
-        }
-        return (URLStreamHandler) addToCache(m_streamHandlerCache, protocol, handler);
-    }
-    
-    private synchronized URLStreamHandler getFromStreamCache(String protocol)
-    {
-        return (URLStreamHandler) ((m_streamHandlerCache != null) ? 
-            m_streamHandlerCache.get(protocol) : null);
-    }
-    
-    private Object addToCache(Map cache, String key, Object value)
-    {
-        if (value == null)
-        {
-            return null;
-        }
-        
-        Object result = cache.get(key);
-            
-        if (result == null)
-        {
-            cache.put(key, value);
-            result = value;
-        }
-        return result;
+        return key != null ? cache.get(key) : null;
     }
 
     /**
@@ -543,8 +593,9 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
      * @param enable a flag indicating whether or not the framework wants to
      *        enable the URL Handlers service.
     **/
-    public static void registerFrameworkInstance(Object framework, boolean enable)
+    public static void registerFrameworkInstance(Felix framework, boolean enable)
     {
+        boolean register = false;
         synchronized (m_frameworks)
         {
             // If the URL Handlers service is not going to be enabled,
@@ -554,13 +605,38 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
                 // We need to create an instance if this is the first
                 // time this method is called, which will set the handler
                 // factories.
-                if (m_handler == null)
+                if (m_handler == null )
                 {
-                    m_handler = new URLHandlers();
+                    register = true;
+                }
+                else
+                {
+                    m_frameworks.add(framework);
+                    m_counter++;
+                }
+            }
+            else
+            {
+                m_counter++;
+            }
+        }
+        if (register)
+        {
+            synchronized (URL.class)
+            {
+                synchronized (m_classloaderToFrameworkLists)
+                {
+                    synchronized (m_frameworks)
+                    {
+                        if (m_handler == null )
+                        {
+                            m_handler = new URLHandlers();
+                        }
+                        m_frameworks.add(framework);
+                        m_counter++;
+                    }
                 }
-                m_frameworks.add(framework);
             }
-            m_counter++;
         }
     }
 
@@ -577,32 +653,55 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
         boolean unregister = false;
         synchronized (m_frameworks)
         {
-            m_counter--;
-            if (m_frameworks.remove(framework))
-            {    
-                if (m_frameworks.isEmpty())
+            if (m_frameworks.contains(framework))
+            {
+                if (m_frameworks.size() == 1 && m_handler != null)
                 {
                     unregister = true;
-                    m_handler = null;
                 }
+                else
+                {
+                    m_frameworks.remove(framework);
+                    m_counter--;
+                }
+            }
+            else
+            {
+                m_counter--;
             }
         }
         if (unregister)
         {
-             try
-             {
-                 m_secureAction.invoke(m_secureAction.getDeclaredMethod(
-                     m_rootURLHandlers.getClass(), 
-                     "unregisterFrameworkListsForContextSearch", 
-                     new Class[]{ ClassLoader.class}), 
-                     m_rootURLHandlers,
-                     new Object[] {URLHANDLERS_CLASS.getClassLoader()});
-             }
-             catch (Exception e)
-             {
-                 // This should not happen
-                 e.printStackTrace();
-             }
+            synchronized (URL.class)
+            {
+                synchronized (m_classloaderToFrameworkLists)
+                {
+                    synchronized (m_frameworks)
+                    {
+                        m_frameworks.remove(framework);
+                        m_counter--;
+                        if (m_frameworks.isEmpty() && m_handler != null)
+                        {
+
+                            m_handler = null;
+                            try
+                            {
+                                m_secureAction.invoke(m_secureAction.getDeclaredMethod(
+                                    m_rootURLHandlers.getClass(),
+                                    "unregisterFrameworkListsForContextSearch",
+                                    new Class[]{ ClassLoader.class}),
+                                    m_rootURLHandlers,
+                                    new Object[] {URLHANDLERS_CLASS.getClassLoader()});
+                            }
+                            catch (Exception e)
+                            {
+                                // This should not happen
+                                e.printStackTrace();
+                            }
+                        }
+                    }
+                }
+            }
         }
     }
 
@@ -619,85 +718,120 @@ class URLHandlers implements URLStreamHandlerFactory, ContentHandlerFactory
     **/
     public static Object getFrameworkFromContext()
     {
-        // This is a hack. The idea is to return the only registered framework
-        synchronized (m_classloaderToFrameworkLists)
+        // This is a hack. The idea is to return the only registered framework quickly
+        int attempts = 0;
+        while (m_classloaderToFrameworkLists.isEmpty() && (m_counter == 1) && (m_frameworks.size() == 1))
         {
-            if (m_classloaderToFrameworkLists.isEmpty())
+            Object framework = m_frameworks.get(0);
+
+            if (framework != null)
             {
-                synchronized (m_frameworks)
-                {
-                    if ((m_counter == 1) && (m_frameworks.size() == 1))
-                    {
-                        return m_frameworks.get(0);
-                    }
-                }
+                return framework;
+            }
+            else if (attempts++ > 3)
+            {
+                break;
             }
         }
+
         // get the current class call stack.
         Class[] stack = m_sm.getClassContext();
         // Find the first class that is loaded from a bundle.
         Class targetClass = null;
+        ClassLoader targetClassLoader = null;
         for (int i = 0; i < stack.length; i++)
         {
-            if (stack[i].getClassLoader() != null) 
+            ClassLoader classLoader = m_secureAction.getClassLoader(stack[i]);
+			if (classLoader != null)
             {
-                String name = stack[i].getClassLoader().getClass().getName();
+                String name = classLoader.getClass().getName();
                 if (name.startsWith("org.apache.felix.framework.ModuleImpl$ModuleClassLoader")
                     || name.equals("org.apache.felix.framework.searchpolicy.ContentClassLoader")
                     || name.startsWith("org.apache.felix.framework.BundleWiringImpl$BundleClassLoader"))
                 {
                     targetClass = stack[i];
+                    targetClassLoader = classLoader;
                     break;
                 }
             }
         }
-        
+
         // If we found a class loaded from a bundle, then iterate
         // over the framework instances and see which framework owns
         // the bundle that loaded the class.
         if (targetClass != null)
         {
-            synchronized (m_classloaderToFrameworkLists)
+            ClassLoader index = m_secureAction.getClassLoader(targetClassLoader.getClass());
+
+            List frameworks = (List) m_classloaderToFrameworkLists.get(index);
+
+            if ((frameworks == null) && (index == URLHANDLERS_CLASS.getClassLoader()))
             {
-                ClassLoader index = targetClass.getClassLoader().getClass().getClassLoader();
-                
-                List frameworks = (List) m_classloaderToFrameworkLists.get(
-                    index);
-                
-                if ((frameworks == null) && (index == URLHANDLERS_CLASS.getClassLoader())) 
+                frameworks = m_frameworks;
+            }
+            if (frameworks != null)
+            {
+                // Check the registry of framework instances
+                for (Object framework : frameworks)
                 {
-                    frameworks = m_frameworks;
-                }
-                if (frameworks != null)
-                {
-                    synchronized (frameworks)
+                    try
                     {
-                        // Check the registry of framework instances
-                        for (int i = 0; i < frameworks.size(); i++)
+                        if (m_secureAction.invoke(
+                            m_secureAction.getDeclaredMethod(framework.getClass(),
+                            "getBundle", CLASS_TYPE),
+                            framework, new Object[]{targetClass}) != null)
                         {
-                            Object framework = frameworks.get(i);
-                            try
-                            {
-                                if (m_secureAction.invoke(
-                                    m_secureAction.getDeclaredMethod(framework.getClass(), 
-                                    "getBundle", CLASS_TYPE),
-                                    framework, new Object[]{targetClass}) != null)
-                                {
-                                    return framework;
-                                }
-                            }
-                            catch (Exception ex)
-                            {
-                                // This should not happen but if it does there is 
-                                // not much we can do other then ignore it.
-                                // Maybe log this or something.
-                                ex.printStackTrace();
-                            }
+                            return framework;
                         }
                     }
+                    catch (Exception ex)
+                    {
+                        // This should not happen but if it does there is
+                        // not much we can do other then ignore it.
+                        // Maybe log this or something.
+                        ex.printStackTrace();
+                    }
                 }
             }
         }
         return null;
     }
+
+    public static Object getFrameworkFromContext(String uuid)
+    {
+        if (uuid != null)
+        {
+            for (Felix framework : m_frameworks)
+            {
+                if (uuid.equals(framework._getProperty(Constants.FRAMEWORK_UUID)))
+                {
+                    return framework;
+                }
+            }
+            for (List frameworks : m_classloaderToFrameworkLists.values())
+            {
+                for (Object framework : frameworks)
+                {
+                    try
+                    {
+                        if (uuid.equals(
+                                m_secureAction.invoke(
+                                        m_secureAction.getDeclaredMethod(framework.getClass(),"getProperty", new Class[]{String.class}),
+                                        framework, new Object[]{Constants.FRAMEWORK_UUID})))
+                        {
+                            return framework;
+                        }
+                    }
+                    catch (Exception ex)
+                    {
+                        // This should not happen but if it does there is
+                        // not much we can do other then ignore it.
+                        // Maybe log this or something.
+                        ex.printStackTrace();
+                    }
+                }
+            }
+        }
+        return getFrameworkFromContext();
+    }
 }
